"""Release [0.83.0].

Revision ID: 0.83.0
Revises: 0.82.1
Create Date: 2025-05-28 08:23:38.860697

"""

# revision identifiers, used by Alembic.
revision = "0.83.0"
down_revision = "0.82.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
