import{aJ as o}from"./@ui-core-DKdciH1Z.js";import{S as r}from"./Tabs-Bxg-GoQW.js";import{R as m}from"./RunsBody-C_OULYXm.js";import{R as i}from"./ButtonGroup-BnqZqOoe.js";import{S as p}from"./StackList-CeeG05KL.js";import{a as s}from"./@react-router-DaIuhn3X.js";import"./index-Bf9OtS7t.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./scrolling-tabs-list-D1IqzDbR.js";import"./transform-Bq7p9csj.js";import"./KeyValue-DanfGmsN.js";import"./NestedCollapsible-Bg3_x-fp.js";import"./eye-off-T0xdenj_.js";import"./eye-0EFEzbgA.js";import"./CollapsibleCard-DT7kno_u.js";import"./CodeSnippet-CT3lMiRZ.js";import"./Pagination-C1_kO1Hb.js";import"./chevron-right-double-CNOR0jlH.js";import"./SearchField-BJOPpUn0.js";import"./all-pipeline-runs-query-DdNqOVs7.js";import"./pipeline-link-BZnbg7wr.js";import"./action-cell-g6pLM9iQ.js";import"./schedule-tag-D_qqKaif.js";import"./clock-DcaJYqTc.js";import"./helper-DWeonard.js";import"./bulk-delete-0CKxjF5U.js";import"./index-Bhbc8WSF.js";import"./stack-detail-query-CHkqdEt4.js";import"./resizable-sheet-BI8dTq_r.js";import"./react-resizable-panels.browser-DzY2Iih7.js";import"./components-Br2ezRib.js";import"./NumberBox-CYok4BWK.js";import"./ComponentBadge-Dd3q4Uyy.js";import"./ComponentIcon-FmTe770m.js";import"./bar-chart-square-check-Lt3fLSJ-.js";import"./layout-BRN9AeDt.js";function M(){const{componentId:t}=s();return o.jsx("div",{className:"@container",children:o.jsx(r,{isPanel:!1,stacksTabContent:o.jsx(p,{fixedQueryParams:{component_id:t}}),runsTabContent:o.jsx(i,{children:o.jsx(m,{fixedQueryParams:{stack_component:t}})}),componentId:t})})}export{M as default};
