import{aJ as t}from"./@ui-core-DKdciH1Z.js";import{C as o,S as r}from"./StackComponentList-DVeGT0JR.js";import"./index-Bf9OtS7t.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./Pagination-C1_kO1Hb.js";import"./chevron-right-double-CNOR0jlH.js";import"./SearchField-BJOPpUn0.js";import"./expand-full-aWqoyakM.js";import"./resizable-sheet-BI8dTq_r.js";import"./react-resizable-panels.browser-DzY2Iih7.js";import"./Tabs-Bxg-GoQW.js";import"./scrolling-tabs-list-D1IqzDbR.js";import"./transform-Bq7p9csj.js";import"./KeyValue-DanfGmsN.js";import"./NestedCollapsible-Bg3_x-fp.js";import"./eye-off-T0xdenj_.js";import"./eye-0EFEzbgA.js";import"./CollapsibleCard-DT7kno_u.js";import"./CodeSnippet-CT3lMiRZ.js";import"./all-pipeline-runs-query-DdNqOVs7.js";import"./index-Bhbc8WSF.js";import"./stack-detail-query-CHkqdEt4.js";import"./ComponentBadge-Dd3q4Uyy.js";import"./ComponentIcon-FmTe770m.js";import"./bar-chart-square-check-Lt3fLSJ-.js";import"./layout-BRN9AeDt.js";import"./action-cell-g6pLM9iQ.js";import"./Infobox-bewh53uP.js";import"./alert-triangle-DWngjzfY.js";import"./helper-DWeonard.js";import"./bulk-delete-0CKxjF5U.js";function F(){return t.jsx("div",{className:"pt-5",children:t.jsx(o,{children:t.jsx(r,{})})})}export{F as default};
