import{aJ as r}from"./@ui-core-DKdciH1Z.js";import{R as o}from"./RunsBody-C_OULYXm.js";import{R as t}from"./ButtonGroup-BnqZqOoe.js";import"./index-Bf9OtS7t.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./Pagination-C1_kO1Hb.js";import"./chevron-right-double-CNOR0jlH.js";import"./SearchField-BJOPpUn0.js";import"./all-pipeline-runs-query-DdNqOVs7.js";import"./pipeline-link-BZnbg7wr.js";import"./action-cell-g6pLM9iQ.js";import"./schedule-tag-D_qqKaif.js";import"./clock-DcaJYqTc.js";import"./helper-DWeonard.js";import"./bulk-delete-0CKxjF5U.js";function v(){return r.jsx(t,{children:r.jsx(o,{})})}export{v as default};
